/**
 * Designs a Howland source based on the input constraints.
 *
 * Usage: ./design < constraints_file
 */

#include <cstdio>
#include <complex>
#include <vector>
#include "opamp.cpp"
#include "data.cpp"
#include "howland.cpp"

extern "C" {
    #include <cpso.h>
}

using namespace std;

const double cost_weight = 1e-9; // higher value tries to save more resources
Constraints cons;
double limits[10], load_min, load_max;
double R = 100, r = 1;
double Vi = (R / r) * 5;

static double fitness(CpsoState* st, const double* v) {
    Howland h( OpAmp(v), R, r);
    
    double mse1 = h.mse_il(cons, Vi, load_min);
    double mse2 = h.mse_il(cons, Vi, load_max);
    
    return -(mse1 + mse2 + h.cost() * cost_weight);
}

#include <sys/time.h>
unsigned long random_seed() {
    struct timeval tv;
    gettimeofday(&tv, 0);
    return tv.tv_usec + tv.tv_sec;
}

int main(int argc, char* argv[]) {
    /* read the limits */
    for(unsigned i = 0; i < 10; ++i)
        scanf("%lf", &limits[i]);

    /* read the target minimum and maximum load */
    scanf("%lf %lf", &load_min, &load_max);

    /* read the constraints */
    cons.load_file(stdin);

    /* configure the PSO algorithm */
    CpsoState* st = cpso_new(5, 40);

    /* configure the limits */
    cpso_set_limits(st, 0, limits[0], limits[1]); // Ri
    cpso_set_limits(st, 1, limits[2], limits[3]); // Ci
    cpso_set_limits(st, 2, limits[4], limits[5]); // Ro
    cpso_set_limits(st, 3, limits[6], limits[7]); // Ao
    cpso_set_limits(st, 4, limits[8], limits[9]); // Wc

    srand( random_seed() );
    cpso_set_eval(st, fitness, 0);
    cpso_set_params(st, 1.85, 1.5, 0.55, 1e-3, 1e-5);

    /* optimize */
    cpso_init(st);
    cpso_run(st, 2000, 0);

    /* output the best solution */
    const double* best;
    double cost = cpso_get_best(st, -1, &best);

    Howland h( OpAmp(best), R, r );
    
    fprintf(stderr, "Computing the total error with the minimum load (%.4f Ohms)\n", load_min);
    double mse_min   = h.mse_il(cons, Vi, load_min, true);
    fprintf(stderr, "Computing the total error with the maximum load (%.4f Ohms)\n", load_max);
    double mse_max   = h.mse_il(cons, Vi, load_max, true);
    double total_mse = mse_min + mse_max;

    fprintf(stderr, "mse (with cost): %g\n", -cost);
    fprintf(stderr, "mse (fit only ): load_min: %g, load_max: %g, total: %g (sqrt(total) = %g)\n", mse_min, mse_max, total_mse, sqrt(total_mse));
    fprintf(stderr, "params:\n");
    for(int i = 0; i < 5; ++i)
        fprintf(stderr, "%12.4lf\n", best[i]);

    vector<double> freqs = logspace(1, 9, 10);
    
    /* load current per frequency */
    freopen("out_il_min", "w", stdout);
    h.freq_sweep_il(freqs, Vi, load_min);
    freopen("out_il_max", "w", stdout);
    h.freq_sweep_il(freqs, Vi, load_max);
    
    /* output current per frequency */
    freopen("out_io", "w", stdout);
    h.freq_sweep_io(freqs, Vi);
    
    /* output impedance per frequency */
    freopen("out_zo", "w", stdout);
    h.freq_sweep(freqs);

    return 0;
}
